/* [KSC4]-------------------------------------------------------------------[]
 *
 *  BCDCLOCK.C - Reduced Pin Count Binary clock for the AVR, version 1.00.000
 *               Written by Marco "Kiko" Carnut <kiko@postcogito.org>
 *
 *  Description: This program counts the time (hours, minutes and seconds),
 *  ------------ displaying it 6-digit or 3-group BCD in either standard
 *  binary order or binary reflected Gray code in a 20-LED display that uses
 *  only 5 pins to drive.
 *
 *  It also sports a primitive keyboard scan/debounce routine for the 3-key
 *  keypad. The first two buttons adjust the time by incrementing hours or
 *  minutes, respectively, and zeroing the seconds counter and stopping the
 *  count. The third button restarts the count if it's stopped (ie., we're in
 *  adjust mode) or cycles the display mode if we're not in adjust mode.
 *
 *  There are 4 display modes: 6-digit BCD in standard binary order, 3-group
 *  BCD in standard binary order, 3-group BCD in Gray code order and 6-digit
 *  BCD in Gray code order.
 *
 *  See also: http://www.postcogito.org/Kiko/BcdClock
 *  ---------
 *
 *  Portability: Tested on an ATtiny45. Compiled under avr-gcc 3.4.4 and
 *  ------------ avr-libc 1.4.0.
 *
 *  License: This program is made available under the Creative Commons
 *  -------- Attribution-NonCommercial-ShareAlike version 2.5 license.
 *  Please see http://creativecommons.org/licenses/by-nc-sa/2.5/ for details.
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * []-----------------------------------------------------------------------[]
 */
 
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>

#define F_CPU          20000000L

#include <util/delay.h>

#ifdef __AVR_ATtiny13__
#define SUPPORTED_CHIP
#endif

#ifdef __AVR_ATtiny25__
#define SUPPORTED_CHIP
#endif

#ifdef __AVR_ATtiny45__
#define SUPPORTED_CHIP
#endif

#ifdef __AVR_ATtiny85__
#define SUPPORTED_CHIP
#endif

#ifndef SUPPORTED_CHIP
#error This MCU is as of yet unsupported
#endif

/* ---------------------- Software Configuration Defines ------------------ */

/* Uncomment this to have all LEDs lit. Useful to test if all LEDs are ok
   and with even brightness. */

//#define HAVE_LAMP_TEST

/* Uncomment this to have all 20 LEDs light up in sequence. Useful to test
   that all connections are ok */
   
//#define HAVE_TEST_MODE

/* Define this if you want to use PB5/RESET to control one of the groups
   of LEDs, leaving PB3 free to use as CLKIN to accept an external clock
   from a crystal oscillator. If you comment this, it will be the /RESET
   pin that will be left being, so you will be able to use SPI-based In-System
   programming, at the expense of having to use the internal RC oscillator,
   which is far too inaccurate for clock applications. */

#define USE_RESET_AS_PBX

/* Move all global variables to registers. The first step to port this program
   to RAMless devices. */

#define USE_REGISTER_VARS

/* ---------------------- Hardware Configuration Defines ------------------ */


#define CNTR_SECONDS           0
#define CNTR_MINUTES           1
#define CNTR_HOURS             2

#define KBD_HOURS              _BV(PB2)
#define KBD_MINUTES            _BV(PB1)
#define KBD_SECONDS            _BV(PB0)

/* --------------------------- Utility Defines ---------------------------- */

#define F_TICK              3125
#define F_CPU_T0_DIV         256
#define PBX_OVERFACTOR_ON      8
#define PBX_OVERFACTOR_OFF     2

#if (F_CPU/F_CPU_T0_DIV)/((F_CPU/F_CPU_T0_DIV/F_TICK-1)+1) != F_TICK
#error Cannot generate exact F_TICK frequency, clock would be inaccurate
#endif

#define KBD_PASSES             8

#define FLAG_BCD               0
#define FLAG_BCD               0
#define FLAG_GRAY              1
#define FLAG_RUNNING           2
#define FLAG_START             3
#define FLAG_TOGGLE_MODE       4

#define TRUE                   1
#define FALSE                  0

#ifdef  USE_RESET_AS_PBX
#define PBX                    PB5
#else
#define PBX                    PB3
#endif // USE_RESET_AS_PBX

#ifdef USE_REGISTER_VARS

#define _r(a) asm(a)

#define DREG                 7
#define DREG_STR             "r7"
#define DREG_PTR             ((unsigned char *)(DREG))

#define KREG                 10
#define KREG_STR             "r10"
#define KREG_PTR             ((unsigned char *)(KREG))

#else

#define register
#define _r(a)
#define DREG_PTR             ((unsigned char *)(&display))
#define KREG_PTR             ((unsigned char *)(&kbd_count))

#endif // USE_REGISTER_VARS

/* --------------------- Global Variables and Declarations ---------------- */

unsigned char pb[] EEMEM = {
    _BV(PB1),_BV(PB0),_BV(PB4),_BV(PBX),_BV(PB2),
    _BV(PB1),_BV(PB0),_BV(PB2),_BV(PB1),_BV(PB4),
    _BV(PB4),_BV(PB0),_BV(PB1),_BV(PBX),_BV(PB2),
    _BV(PB4),_BV(PB0),_BV(PBX),_BV(PB2),_BV(PBX)
};

unsigned char db[] EEMEM = {
    _BV(PB1)|_BV(PB2),_BV(PB0)|_BV(PB1),_BV(PB4)|_BV(PB0),_BV(PBX)|_BV(PB4),
    _BV(PB2)|_BV(PB1),_BV(PB1)|_BV(PB0),_BV(PB0)|_BV(PB4),_BV(PB2)|_BV(PB0),
    _BV(PB1)|_BV(PB4),_BV(PB4)|_BV(PB1),_BV(PB4)|_BV(PBX),_BV(PB0)|_BV(PB2),
    _BV(PB1)|_BV(PBX),_BV(PBX)|_BV(PB1),_BV(PB2)|_BV(PB4),_BV(PB4)|_BV(PB2),
    _BV(PB0)|_BV(PBX),_BV(PBX)|_BV(PB0),_BV(PB2)|_BV(PBX),_BV(PBX)|_BV(PB2)
};

register unsigned char  cycle      _r("r2");
register unsigned char  cntr       _r("r3");
register unsigned short ticks      _r("r4");

register struct {
    unsigned char passes:4;
    unsigned char ocntr:4;
} packed _r("r6");

#ifdef USE_REGISTER_VARS

register unsigned char display    _r(DREG_STR);
register unsigned char kbd_count  _r(KREG_STR);

#else

unsigned char display[3];
unsigned char kbd_count[3];

#endif // USE_REGISTER_VARS

#ifdef __AVR_ATtiny13__

register unsigned char flags _r("r13");

#define GPIOR1 flags
#define TIMSK  TIMSK0

#endif // __AVR_ATtiny13__


/* --------------------------- Function Prototypes ------------------------ */

unsigned char count_up(unsigned char i);
unsigned char xform(unsigned char value);

/* ------------------------------ Main Program ---------------------------- */

int main(void)
{
    /* Timer0 Setup. We will get an interrupt F_TICK times per second.
       The interrupt is used just to wake the part from idle mode, so
       the interrupt handler is empty. */
    TCCR0A = _BV(WGM01);          // CTC mode, no PWM
#if   F_CPU_T0_DIV == 64
    TCCR0B = 3;
#elif F_CPU_T0_DIV == 256
    TCCR0B = 4;
#else
#error Unknown F_CPU_T0_DIV
#endif // F_CPU_T0_DIV
    OCR0A  = F_CPU/F_CPU_T0_DIV/F_TICK-1;
    TIMSK  = _BV(OCIE0A);

    sei();
    packed.ocntr=1;
    
#ifdef USE_REGISTER_VARS
    DREG_PTR[0]=DREG_PTR[1]=DREG_PTR[2]=0;
#endif // USE_REGISTER_VARS

#ifdef HAVE_TEST_MODE
    DREG_PTR[0]=1;
#endif // HAVE_TEST_MODE

    /* We start in BCD mode and with the clock running */
    GPIOR1 = _BV(FLAG_RUNNING)|_BV(FLAG_BCD);
    for (;;) {
        if (!--packed.ocntr) {
            if (cycle==20) {              // Handle keypad
                DDRB = PORTB = 0;         // PORTB is all inputs, no pullups
                _delay_us(50);            // Wait for ports to settle
                unsigned char n;
                unsigned char k=PINB;    // Get the keys statuses
                // Primitive debounce routine: count how many times it's on
                for (n=0;n<3;n++) {
                    if (k&1) KREG_PTR[n]++;
                    k>>=1;
                }
                // If we reached the number of scan packed.passes, let's check
                // the results
                if (++packed.passes>=KBD_PASSES) {
                    packed.passes=0;
                    unsigned char out=0;
                    for (n=0;n<3;n++) {
                        out>>=1;
                        if (KREG_PTR[n]>KBD_PASSES/2) out|=4;
                        KREG_PTR[n]=0;
                    }
                    if (!out) {
                        // Commands that need to be executed when all keys
                        // are released
                        if (GPIOR1 & _BV(FLAG_TOGGLE_MODE)) {
                            GPIOR1 &= ~_BV(FLAG_TOGGLE_MODE);
                            unsigned char mode=GPIOR1 & 3;
                            if (mode==_BV(FLAG_BCD)) {
                                GPIOR1 &= ~_BV(FLAG_BCD);
                            } else if (mode==0) {
                                GPIOR1 |= _BV(FLAG_GRAY);
                            } else if (mode==_BV(FLAG_GRAY)) {
                                GPIOR1 |= _BV(FLAG_BCD);
                            } else GPIOR1 &= ~_BV(FLAG_GRAY);
                        }
                        if (GPIOR1 & _BV(FLAG_START)) {
                            GPIOR1 &= ~_BV(FLAG_START);
                            GPIOR1 |= _BV(FLAG_RUNNING);
                        }
                    } else if (out==KBD_HOURS) {
                        GPIOR1 &= ~_BV(FLAG_RUNNING);
                        DREG_PTR[CNTR_SECONDS]=0;
                        count_up(CNTR_HOURS);
                    } else if (out==KBD_MINUTES) {
                        GPIOR1 &= ~_BV(FLAG_RUNNING);
                        DREG_PTR[CNTR_SECONDS]=0;
                        count_up(CNTR_MINUTES);
                    } else if (out==KBD_SECONDS) {
                        if (GPIOR1 & _BV(FLAG_RUNNING)) {
                            GPIOR1 |= _BV(FLAG_TOGGLE_MODE);
                        } else {
                            GPIOR1 |= _BV(FLAG_START);
                        }
                    }
                }
                packed.ocntr = 1;
            } else {                      // Handle display
                if (cycle==7) {
                    cntr=xform(DREG_PTR[1]);
                } else if (cycle==14) {
                    cntr=xform(DREG_PTR[2]);
                }
                unsigned char p,d;
                p=eeprom_read_byte(pb+cycle);
                d=eeprom_read_byte(db+cycle);
                if (d & _BV(PBX)) {
                    if (p & _BV(PBX)) {
                        packed.ocntr = PBX_OVERFACTOR_ON;
                    } else {
                        packed.ocntr = PBX_OVERFACTOR_OFF;
                    }
                } else {
                    packed.ocntr = 1;
                }
                if (cntr&1) {
                    PORTB = p;
                    DDRB  = d;
                } else {
                    DDRB = PORTB = 0;
                }
                cntr>>=1;
            }
            if (++cycle==21) {
                cycle=0;
                cntr=xform(DREG_PTR[0]);
            }
        }
        if (++ticks==F_TICK) {
            ticks=0;
#ifdef HAVE_TEST_MODE
            if (DREG_PTR[0]) {
                DREG_PTR[0]<<=1;
                if (DREG_PTR[0]==0x80) {
                    DREG_PTR[0]=0; DREG_PTR[1]=1;
                }
            } else if (DREG_PTR[1]) {
                DREG_PTR[1]<<=1;
                if (DREG_PTR[1]==0x80) {
                    DREG_PTR[1]=0; DREG_PTR[2]=1;
                }
            } else {
                DREG_PTR[2]<<=1;
                if (DREG_PTR[2]==0x40) {
                    DREG_PTR[2]=0; DREG_PTR[0]=1;
                }
            }
#else
            if (GPIOR1 & _BV(FLAG_RUNNING)) {
                if (count_up(CNTR_SECONDS))
                    if (count_up(CNTR_MINUTES))
                        count_up(CNTR_HOURS);
            }
#endif // HAVE_TEST_MODE
        }
        set_sleep_mode(SLEEP_MODE_IDLE);
        sleep_mode();
        asm("nop");
    }
    return 0;
}

/* Increments counter 'i' in BCD mode, wrapping at the proper limit (24 for
   hours, 60 for minutes and seconds). Returns TRUE if there was a carry,
   false otherwise. */

unsigned char count_up(unsigned char i)
{
    unsigned char limit = (i==2) ? 0x23 : 0x59;
    if (DREG_PTR[i]>=limit) {
        DREG_PTR[i]=0;
        return TRUE;
    } 
    if ((DREG_PTR[i]&15)>=9)
        DREG_PTR[i]+=7;
    else 
        DREG_PTR[i]++;
    return FALSE;
}

unsigned char xform(unsigned char value)
{
#ifdef HAVE_LAMP_TEST
    return 0x7F;
#endif // HAVE_LAMP_TEST
    if (GPIOR1 & _BV(FLAG_BCD)) {
        /* If we're in BCD mode. If we're not in Gray mode, there's nothing
           to do -- we just return with no further action. */
        if (!(GPIOR1 & _BV(FLAG_GRAY))) return value;
        /* The code below separates each nibble, converts them to Gray code
           and joins them back. */
        unsigned char lo=value&15;
        unsigned char hi=value>>4;
        hi^=(hi>>1);
        lo^=(lo>>1);
        return (hi<<4)|lo;
    } else {
        /* We're in pure binary mode, so we need to convert from BCD to pure
           binary. The code below does this by computing 10*hi+lo, where 'hi'
           is the most significant nibble and 'lo' is the least siginificant
           nibble. */
        unsigned char lo=value&15;
        unsigned char h1=(value&0xF0)>>1;
        unsigned char h2=h1>>2;
        value=h1+h2+lo;
        /* If we're not using Gray code, return the value as is */
        if (!(GPIOR1 & _BV(FLAG_GRAY))) return value;
        /* Otherwise, convert to Gray code */
        value^=value>>1;
        return value;        
    }
}

/* ---------------------------- Interrupt Handlers ------------------------ */

/* Empty interrupt handler for the Output Compare 0A source. This is interrupt
   exists just to wake the part from idle sleep mode. */

EMPTY_INTERRUPT(SIG_OUTPUT_COMPARE0A);
